#include "stdlib.h"
#include "realtime_threads.h"
#include "async_event_handler.h"
#include "threads.h"
#include "exceptions.h"
#include "interpreter.h"
#include "specialsignatures.h"
#include "stack.h"
#include "configure.h"




/**
 * List of waitting start time Thread
 * 
 */
static RealtimeThread *rtThreadQ;




void init_rt_threads(void){
	rtThreadQ = NULL;
}

void init_rt_thread(RealtimeThread *rtThread, byte memoryType){


#if IMMORTAL_MEM_ACTIVED == 0
	if(memoryType == IMMORTAL_MEM) {
		throw_exception(illegalStateException);
		return;
	}
#endif


	//protect_obj(rtThread);
	int now = (int)get_sys_time();

	rtThread->memoryType = memoryType;

	init_thread(rt_get_thread(rtThread));
	
	if(rtThread->start > 0){
		rtThread->startPeriod = now + rtThread->start; 
		// we can't use sleep_thread because our thread
		// is not the currentThread.
		rt_get_thread(rtThread)->sleepUntil = rtThread->start + get_sys_time();
		rt_get_thread(rtThread)->state = START_DELAYED;
	} else {
		rtThread->startPeriod = now; 
	}
	rtThread->endDeadline = rtThread->startPeriod + rtThread->deadline; 
	rtThread->remainingCost = rtThread->cost; 


	//add realtime thread into realtime thread queue
	rtThread->nextRealtimeThread = ptr2ref(rtThreadQ);
	rtThreadQ = rtThread;
	//unprotect_obj(rtThread);
}

void prepare_2_wait(RealtimeThread *rtThread){
	FOURBYTES time = (rtThread->startPeriod + rtThread->period);

	//set up the time of the next start
	if(time > get_sys_time()){
		// we need to use get_sys_time() and not 
		// a variable sets earlier to be more precise.
		sleep_thread(time - get_sys_time());
	} else  {
		//rt_get_thread(rtThread)->sleepUntil = 0;
	}
	
	//update the absolute time of the realtime thread
	rtThread->startPeriod += rtThread->period;
	rtThread->endDeadline += rtThread->period;
	
	rtThread->remainingCost = rtThread->cost; 


	schedule_request(REQUEST_SWITCH_THREAD);
}

inline void rt_set_start(RealtimeThread *rtThread, JINT start) {
	rtThread->start = start;
}

inline void rt_set_period(RealtimeThread *rtThread, JINT period) {
	rtThread->period = period;
}

inline void rt_set_cost(RealtimeThread *rtThread, int cost) {
	rtThread->cost = cost;
}

inline void rt_set_deadline(RealtimeThread *rtThread, int deadline) {
	rtThread->deadline = deadline;
	
}

void check_deadline() {
	RealtimeThread *rtthread = rtThreadQ;
	int now = (int)get_sys_time();
//	protect_obj(rtThreadQ);
	while(rtthread){
		//check deadline
		//deadline == 0  infinite time
		if(rtthread->deadline != 0 && rtthread->endDeadline < now){

			if(rtthread->deadlineMissHandler){
				getAndIncrementPendingFireCount(ref2ptr(rtthread->deadlineMissHandler));
			}
			dequeue_rt_thread(rtthread);	
			dequeue_thread(rt_get_thread(rtthread));
			schedule_request(REQUEST_SWITCH_THREAD);

			//lancement du handler 
/*			set_top_ref_cur (ptr2ref ());
			dispatch_virtual ((Object *) candidate, run_4_5V, null);
*/		}
		

		//check cost overrun on the current thread
		// cost ==0 infinite time
		if(currentThread->threadId == rt_get_thread(rtthread)->threadId && rtthread->cost != 0){
			//rtthread->remainingCost -= (now - thread_start_time);
			rtthread->remainingCost -= (get_sys_time() - thread_start_time);
			if(rtthread->remainingCost < 0){

				if(rtthread->costOverrunHandler){
					getAndIncrementPendingFireCount(ref2ptr(rtthread->costOverrunHandler));
				}
				dequeue_rt_thread(rtthread);	
				dequeue_thread(rt_get_thread(rtthread));
				schedule_request(REQUEST_SWITCH_THREAD);
			}
		}

		rtthread = rtthread->nextRealtimeThread;
	}
//unprotect_obj(rtThreadQ);
}

void dequeue_rt_thread(RealtimeThread *rtthread){
	RealtimeThread *rtt = rtThreadQ;
	if(rtthread == NULL) return;
	if(rtt == rtthread){
		rtThreadQ = ref2ptr(rtThreadQ->nextRealtimeThread);
	} else {
		while((rtt != NULL) && ((rtt = rtt->nextRealtimeThread) != rtthread));
		if(rtt != NULL)
			rtt->nextRealtimeThread = rtthread->nextRealtimeThread;
	}
	rtthread->nextRealtimeThread = NULL;
}

RealtimeThread* current_is_rt(){
	 RealtimeThread *rtt = rtThreadQ;
    while(rtt){
			if(rt_get_thread(rtt)->threadId == currentThread->threadId){
				return rtt;
			}
			rtt = rtt->nextRealtimeThread;
	  }
	  return NULL;
}

inline void schedule_periodic(RealtimeThread *rtthread){		
	//on calcul la prochaine activation
	// et les autres temps
	rtthread->startPeriod += rtthread->period;
	rtthread->endDeadline += rtthread->period;

//	printf("time: %d, next start: %d, next dl: %d\n", now, rtthread->startPeriod, rtthread->endDeadline);

	//on le rajoute dans la liste de rt Thread
	rtthread->nextRealtimeThread = rtThreadQ;
	rtThreadQ = rtthread;
	//on rajoute dans le scheduleur
	enqueue_thread(rt_get_thread(rtthread));	
}

RealtimeThread* findRealtimeThread(Thread * t){
  RealtimeThread* anchorRtThread = rtThreadQ;
  if (anchorRtThread == null){
    return null;
  }

  RealtimeThread* currentRtThread = anchorRtThread;
  do{
    if (rt_get_thread(currentRtThread) == t){
      return currentRtThread;
    }
    currentRtThread = currentRtThread->nextRealtimeThread;

  }while (currentRtThread != null && currentRtThread != anchorRtThread);

  return null;
}
